"""Methods for reading and writing user configuration file settings.

"""
import os
from base64 import urlsafe_b64decode, urlsafe_b64encode


APPDIR = os.getenv('APPDIR', '/home/fieldedge/fieldedge')
USERDIR = os.getenv('USERDIR', f'{APPDIR}/user')
USER_CONFIG_FILE = os.getenv('USER_CONFIG_FILE', f'{USERDIR}/config.env')


def read_user_config(filename: str = USER_CONFIG_FILE) -> dict:
    """Reads user configuration from a `.env` style file.
    
    Format of the file is `CONST_CASE=value` with one entry per line.
    
    Args:
        filename: The full path/filename
    
    Returns:
        A dictionary of configuration settings.
        
    """
    user_config = {}
    if not isinstance(filename, str):
        filename = USER_CONFIG_FILE
    if os.path.isfile(filename):
        with open(filename) as file:
            for line in file.readlines():
                if line.startswith('#') or not line.strip():
                    continue
                key, value = line.split('=', 1)
                user_config[key] = value.strip()
                if 'PASSWORD' in key:
                    user_config[key] = unobscure(value)
    return user_config


def write_user_config(config: dict, filename: str = USER_CONFIG_FILE) -> None:
    """Writes the user config values to the file path specified.
    
    Keys are converted to `CONST_CASE`.
    
    Args:
        config: The configuration settings dictionary.
        filename: The full file path/name to store into. Defaults to
            `{APPDIR}/user/config.env`, where `APPDIR` is an environment
            variable with default `/home/fieldedge/fieldedge`.
        
    """
    lines_to_write: 'list[str]' = []
    keys_written: 'list[str]' = []
    if not isinstance(filename, str) or not os.path.dirname(filename):
        filename = USER_CONFIG_FILE
    if os.path.isfile(filename):
        with open(filename) as file:
            for line in file.readlines():
                if line.startswith('#') or not line.strip():
                    continue
                file_key, file_value = line.strip().split('=', 1)
                if file_key in config:
                    keys_written.append(file_key)
                    if 'PASSWORD' in file_key:
                        file_value = unobscure(file_value)
                    if file_value != config[file_key]:
                        new_value = config[file_key]
                        if 'PASSWORD' in file_key:
                            new_value = obscure(new_value)
                        lines_to_write.append(f'{file_key}={new_value}')
                        continue
                lines_to_write.append(line.strip())
    for key, val in config.items():
        if key in keys_written:
            continue
        if 'PASSWORD' in key:
            lines_to_write.append(f'{key}={obscure(val)}')
        else:
            lines_to_write.append(f'{key}={val}')
    with open(filename, 'w') as file:
        file.writelines('\n'.join(lines_to_write))


def obscure(value: str) -> str:
    """Obscures a value for simple security."""
    return urlsafe_b64encode(value.encode()).decode()


def unobscure(obscured: str) -> str:
    """Unobscures a value previously obscured."""
    return urlsafe_b64decode(obscured.encode()).decode()
     


"""A proxy class for interfacing with other Microservices via MQTT.
"""
import logging
from typing import Callable

from fieldedge_utilities.mqtt import MqttClient

__all__ = ['SubscriptionProxy']

_log = logging.getLogger(__name__)


class SubscriptionProxy:
    """Passes MQTT topic/message from a parent to a child object.
    
    For example a MicoserviceProxy may want to listen for other microservice
    events than just the one it is a proxy for.
    
    """

    def __init__(self, mqtt_client: MqttClient) -> None:
        """Initializes the subscription proxy.
        
        Args:
            mqtt_client (MqttClient): The parent MQTT client.
            
        """
        if not isinstance(mqtt_client, MqttClient):
            raise ValueError('mqtt_client must be a valid MqttClient instance')
        self._mqttc: MqttClient = mqtt_client
        self._subscriptions: dict = {}

    def proxy_add(self,
                  module: str,
                  topic: str,
                  callback: Callable,
                  qos: int = 0) -> bool:
        """Adds a subscription proxy to the parent.
        
        Args:
            module: The module name used as a routing key.
            topic: The MQTT topic e.g. `fieldedge/my-microservice/events/#`
            callback: The callback function that will receive the MQTT publish
                `(topic: str, message: dict)`
            qos: The MQTT QoS 0 = max once, 1 = at least once, 2 = exactly once
            
        """
        for mod, topics in self._subscriptions.items():
            for top in topics:
                if top == topic and mod == module:
                    _log.warning('Topic %s already subscribed by %s',
                                 topic, mod)
                    return False
        if module not in self._subscriptions:
            self._subscriptions[module] = {}
        try:
            self._mqttc.subscribe(topic, qos)
            self._subscriptions[module][topic] = callback
            return True
        except Exception as err:
            _log.error('Failed to proxy subscribe: %s', err)
            return False

    def proxy_del(self, module: str, topic: str) -> bool:
        """Removes a subscription proxy."""
        modules_subscribed = []
        for mod, topics in self._subscriptions.items():
            for top in topics:
                if top == topic:
                    modules_subscribed.append(mod)
        if (module in self._subscriptions and
            topic in self._subscriptions[module]):
            # found it - ok to remove
            try:
                del self._subscriptions[module][topic]
                if not self._subscriptions[module]:
                    del self._subscriptions[module]
                if len(modules_subscribed) == 1:
                    self._mqttc.unsubscribe(topic)
                return True
            except Exception as err:
                _log.error('Failed to proxy unsubscribe: %s', err)
                return False
        return True

    def proxy_pub(self, topic: str, message: dict) -> None:
        """Publishes via a parent MQTT publish function."""
        for topics in self._subscriptions.values():
            if topic in topics:
                if callable(topics[topic]):
                    topics[topic](topic, message)



"""A Feature class for use as a child of a `Microservice`.
"""
from abc import ABC, abstractmethod
from typing import Any, Callable

from fieldedge_utilities.properties import camel_case, get_class_properties

from .interservice import IscTaskQueue

__all__ = ['Feature']


class Feature(ABC):
    """Template for a microservice feature as a child of the microservice.
    
    References the parent microservice's IscTaskQueue and methods to callback
    for task notification/complete/fail as private attributes.
    
    """

    __slots__ = ['_task_queue', '_task_notify', '_task_complete', '_task_fail']

    def __init__(self,
                 task_queue: IscTaskQueue = None,
                 task_notify: Callable[[str, dict], None] = None,
                 task_complete: Callable[[str, dict], None] = None,
                 task_fail: Callable[[Any], None] = None,
                 **kwargs) -> None:
        """Initializes the feature.
        
        Args:
            task_queue (`IscTaskQueue`): The parent microservice ISC task queue.
            task_notify (`Callable[[str, dict]]`): The parent `notify`
                method for MQTT publish.
            task_complete (`Callable[[str, dict]]`): A parent task
                completion function to receive task `uid` and `task_meta`.
            task_fail (`Callable`): An optional parent function to call if the
                task fails.

        Keyword Args:
             
        """
        self._task_queue: IscTaskQueue = task_queue
        self._task_notify: Callable[[str, dict], None] = task_notify
        self._task_complete: Callable[[str, dict], None] = task_complete
        self._task_fail: Callable[[Any], None] = task_fail
        for key, val in kwargs.items():
            self.__slots__.append(key)
            setattr(self, key, val)

    @property
    def tag(self) -> str:
        try:
            return getattr(self, '_tag')
        except AttributeError:
            return self.__class__.__name__.lower()

    @abstractmethod
    def properties_list(self) -> 'list[str]':
        """Returns a lists of exposed property names."""
        return get_class_properties(self)

    @abstractmethod
    def status(self) -> dict:
        """Returns a dictionary of key status summary information."""
        return {camel_case(key): getattr(self, key)
                for key in self.properties_list}

    @abstractmethod
    def on_isc_message(self, topic: str, message: dict) -> bool:
        """Called by a parent Microservice to pass relevant MQTT messages.
        
        Args:
            topic (str): The message topic.
            message (dict): The message content.
        
        Returns:
            `True` if the message was processed or `False` otherwise.
            
        """
        return False



"""
.. include:: ../../docs/overviews/microservice.md
"""
from .feature import Feature
from .interservice import IscException, IscTask, IscTaskQueue
from .microservice import Microservice, QueuedCallback
from .msproxy import MicroserviceProxy
from .propertycache import PropertyCache
from .subscriptionproxy import SubscriptionProxy

__all__ = [
    'Feature',
    'IscException',
    'IscTask',
    'IscTaskQueue',
    'Microservice',
    'QueuedCallback',
    'MicroserviceProxy',
    'PropertyCache', 
    'SubscriptionProxy',
]
